searchData={"items":[{"type":"module","title":"InstructorLite","doc":"Main building blocks of InstructorLite.","ref":"InstructorLite.html"},{"type":"module","title":"Key Concepts - InstructorLite","doc":"Structured prompting can be quite different depending on the LLM and InstructorLite does only the bare minimum to abstract this complexity. This means the usage can be quite different depending on the adapter you're using, so make sure to consult adapter documentation to learn the details.\n\nThere are two key arguments used throughout this module. Understanding what they are will make your life a lot easier.\n\n* `params` - is an adapter-specific map, that contain values eventually sent to the LLM. More simply, this is the body that will be posted to the API endpoint. You prompt, model name, optional parameters like temperature all likely belong here.\n* `opts` - is a list of options that shape behavior of InstructorLite itself. Options may include things like which schema to cast response to, http client to use, api key, optional headers, http timeout, etc.","ref":"InstructorLite.html#module-key-concepts"},{"type":"module","title":"Shared options - InstructorLite","doc":"Most functions in this module accept a list of options.\n* `:response_model` - Required. A module implementing `InstructorLite.Instruction` behaviour, Ecto schema or [schemaless Ecto definition](https://hexdocs.pm/ecto/Ecto.Changeset.html#module-schemaless-changesets)\n\n* `:adapter` (`t:atom/0`) - A module implementing `InstructorLite.Adapter` behaviour. The default value is `InstructorLite.Adapters.OpenAI`.\n\n* `:max_retries` (`t:non_neg_integer/0`) - How many additional attempts to make if changeset validation fails. The default value is `0`.\n\n* `:validate_changeset` (function of arity 2) - Override function to be called instead of `response_model.validate_changeset/2` callback\n\n* `:notes` (`t:String.t/0`) - Additional notes about the schema that might be used by an adapter\n\n* `:json_schema` (`t:map/0`) - JSON schema to use instead of calling response_model.json_schema/0 callback or generating it at runtime using `InstructorLite.JSONSchema` module\n\n* `:adapter_context` (`t:term/0`) - Options used by adapter callbacks. See adapter docs for schema.\n\n* `:extra` (`t:term/0`) - Any arbitrary term for ad-hoc usage. For example, in `c:InstructorLite.Instruction.validate_changeset/2` callback","ref":"InstructorLite.html#module-shared-options"},{"type":"function","title":"InstructorLite.consume_response/3","doc":"Triage raw LLM response\n\nAttempts to cast raw response from `c:InstructorLite.Adapter.send_request/2` and either returns an object or an invalid changeset with new prompt that can be used for a retry.\n\nThis function will call `c:InstructorLite.Instruction.validate_changeset/2` callback, unless `validate_changeset` option is overridden in `opts`.","ref":"InstructorLite.html#consume_response/3"},{"type":"function","title":"InstructorLite.instruct/2","doc":"Perform instruction session from start to finish.\n\nThis function glues together all other functions and adds retries on top.","ref":"InstructorLite.html#instruct/2"},{"type":"function","title":"Examples - InstructorLite.instruct/2","doc":"","ref":"InstructorLite.html#instruct/2-examples"},{"type":"function","title":"Basic Example - InstructorLite.instruct/2","doc":"<!-- tabs-open -->","ref":"InstructorLite.html#instruct/2-basic-example"},{"type":"function","title":"OpenAI - InstructorLite.instruct/2","doc":"```\niex> InstructorLite.instruct(%{\n    input: [\n      %{role: \"user\", content: \"John Doe is fourty two years old\"}\n    ]\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.OpenAI,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```","ref":"InstructorLite.html#instruct/2-openai"},{"type":"function","title":"Anthropic - InstructorLite.instruct/2","doc":"```\niex> InstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"John Doe is fourty two years old\"}\n    ]\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Anthropic,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :anthropic_key)]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```","ref":"InstructorLite.html#instruct/2-anthropic"},{"type":"function","title":"Llamacpp - InstructorLite.instruct/2","doc":"```\niex> InstructorLite.instruct(%{\n    prompt: \"John Doe is fourty two years old\"\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Llamacpp,\n  adapter_context: [url: Application.fetch_env!(:instructor_lite, :llamacpp_url)]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```","ref":"InstructorLite.html#instruct/2-llamacpp"},{"type":"function","title":"Gemini - InstructorLite.instruct/2","doc":"```elixir\niex> InstructorLite.instruct(%{\n    contents: [\n      %{\n        role: \"user\",\n        parts: [%{text: \"John Doe is fourty two years old\"}]\n      }\n    ]\n  },\n  response_model: %{name: :string, age: :integer},\n  json_schema: %{\n    type: \"object\",\n    required: [:age, :name],\n    properties: %{name: %{type: \"string\"}, age: %{type: \"integer\"}}\n  },\n  adapter: InstructorLite.Adapters.Gemini,\n  adapter_context: [\n    api_key: Application.fetch_env!(:instructor_lite, :gemini_key)\n  ]\n)\n{:ok, %{name: \"John Doe\", age: 42}}\n```\n\n<!-- tabs-close -->","ref":"InstructorLite.html#instruct/2-gemini"},{"type":"function","title":"Using max_retries - InstructorLite.instruct/2","doc":"```\ndefmodule Rhymes do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n  \n  @primary_key false\n  embedded_schema do\n    field(:word, :string)\n    field(:rhymes, {:array, :string})\n  end\n  \n  @impl true\n  def validate_changeset(changeset, _opts) do\n    Ecto.Changeset.validate_length(changeset, :rhymes, is: 3)\n  end\nend\n\nInstructorLite.instruct(%{\n    input: [\n      %{role: \"user\", content: \"Take the last word from the following line and add some rhymes to it\nEven though you broke my heart\"}\n    ]\n  },\n  response_model: Rhymes,\n  max_retries: 1,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %Rhymes{word: \"heart\", rhymes: [\"part\", \"start\", \"dart\"]}}\n```","ref":"InstructorLite.html#instruct/2-using-max_retries"},{"type":"type","title":"InstructorLite.opts/0","doc":"Options passed to instructor functions.","ref":"InstructorLite.html#t:opts/0"},{"type":"function","title":"InstructorLite.prepare_prompt/2","doc":"Prepare prompt that can be later sent to LLM\n\nThe prompt is added to `params`, so you need to cooperate with the adapter to know what you can provide there.\n\nThe function will call `c:InstructorLite.Instruction.notes/0` and `c:InstructorLite.Instruction.json_schema/0` callbacks for `response_model`. Both can be overriden with corresponding options in `opts`.","ref":"InstructorLite.html#prepare_prompt/2"},{"type":"module","title":"InstructorLite.JSONSchema","doc":"Helper module to generate JSON Schema based on Ecto schema.\n\nJSON Schema comes in many flavors and different LLMs have different limitations. Currently, this module aims to implement a schema suitable for [OpenAI structured outputs](https://platform.openai.com/docs/guides/structured-outputs/supported-schemas). This type of JSON schema may not be optimal or compatible with other models. Therefore, it's recommended to use this module as a starting point to generate a schema during development. You can then manually refine the schema and either implement it in the `c:InstructorLite.Instruction.json_schema/0` callback or pass it as an option.\n\n> #### Warning {: .warning}\n>\n> For the reasons described above, neither backward compatibility nor compatibility with all built-in adapters are goals of this module.","ref":"InstructorLite.JSONSchema.html"},{"type":"function","title":"InstructorLite.JSONSchema.from_ecto_schema/1","doc":"Generates a JSON Schema from an Ecto schema.","ref":"InstructorLite.JSONSchema.html#from_ecto_schema/1"},{"type":"behaviour","title":"InstructorLite.Adapter","doc":"Behaviour for implementing adapter modules.\n\nThe role of an adapter module is to encapsulate all logic for a particular LLM, which can be quite specific. As a result, most of the details live in adapter modules and main InstructorLite interface is very broad.\n  \n> #### Implement your own! {: .tip}\n>\n> While built-in adapters are fairly flexible, users are encouraged to write their own adapters to establish a better control over prompt building, http clients, etc.","ref":"InstructorLite.Adapter.html"},{"type":"callback","title":"InstructorLite.Adapter.initial_prompt/2","doc":"Update `params` with model-specific prompt.","ref":"InstructorLite.Adapter.html#c:initial_prompt/2"},{"type":"type","title":"InstructorLite.Adapter.params/0","doc":"Map of adapter-specific values, such as messages, prompt, model name, or temperature. These parameters are typically sent to the LLM in request body.","ref":"InstructorLite.Adapter.html#t:params/0"},{"type":"callback","title":"InstructorLite.Adapter.parse_response/2","doc":"Parse API response.\n\nOn success, it returns an ok tuple with parsed json object ready to be cast to a changeset.\n\nOn error, it may return a 3-element error tuple with error reason. In the worst-case scenario, it returns the original input wrapped in an error tuple.","ref":"InstructorLite.Adapter.html#c:parse_response/2"},{"type":"type","title":"InstructorLite.Adapter.parsed_response/0","doc":"Parsed response content that can be cast to a changeset.","ref":"InstructorLite.Adapter.html#t:parsed_response/0"},{"type":"type","title":"InstructorLite.Adapter.response/0","doc":"Raw content of a successful response.","ref":"InstructorLite.Adapter.html#t:response/0"},{"type":"callback","title":"InstructorLite.Adapter.retry_prompt/5","doc":"Update `params` with model-specific prompt for a retry. This usually involves appending errors from a changeset.","ref":"InstructorLite.Adapter.html#c:retry_prompt/5"},{"type":"callback","title":"InstructorLite.Adapter.send_request/2","doc":"Make request to model API.","ref":"InstructorLite.Adapter.html#c:send_request/2"},{"type":"behaviour","title":"InstructorLite.Instruction","doc":"`use InstructorLite.Instruction` is a way to make your Ecto schema into an Instruction, which provides additional callbacks used by InstructorLite.","ref":"InstructorLite.Instruction.html"},{"type":"behaviour","title":"Example - InstructorLite.Instruction","doc":"```\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  Field Descriptions:\n  - class: Whether or not the email is spam.\n  - reason: A short, less than 10-word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl InstructorLite.Instruction\n  def validate_changeset(changeset, _opts) do\n    Ecto.Changeset.validate_number(changeset, :score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n  \n> #### `use InstructorLite.Instruction` {: .info}\n>\n> When you `use InstructorLite.Instruction`, the Instruction module will set `@behaviour InstructorLite.Instruction`, and define default implementations of `c:notes/0` and `c:json_schema/0` callbacks.","ref":"InstructorLite.Instruction.html#module-example"},{"type":"callback","title":"InstructorLite.Instruction.json_schema/0","doc":"Defines JSON schema for the instruction.\n  \nBy default, `InstructorLite.JSONSchema.from_ecto_schema/1` is called at runtime every time InstructorLite needs to convert an Ecto schema to JSON schema. However, you can bake your own JSON schema into the `c:json_schema/0` callback to eliminate the need to do it on every call.\n\n> #### Tip {: .tip}\n> \n> Take advantage of this callback! Most JSON schemas are known ahead of time, so there is no need to constantly build them at runtime. In addition, `InstructorLite.JSONSchema` module aims to generate one-size-fits-all schema, so it's very unlikely to take full advantage of JSON capabilities of your LLM of choice.","ref":"InstructorLite.Instruction.html#c:json_schema/0"},{"type":"callback","title":"InstructorLite.Instruction.notes/0","doc":"Defines an optional free-form description of the schema.\n  \nYou can define a `@notes` attribute or `c:notes/0` callback to provide an additional free-form description to the schema. This description will be passed to the LLM together with the schema definition.","ref":"InstructorLite.Instruction.html#c:notes/0"},{"type":"callback","title":"InstructorLite.Instruction.validate_changeset/2","doc":"Called by `InstructorLite.consume_response/3` as part of changeset validation.\n\nIt has full access to all `opts`. If you need to pass an arbitrary term to this callback, use the `extra` key.","ref":"InstructorLite.Instruction.html#c:validate_changeset/2"},{"type":"callback","title":"Example - InstructorLite.Instruction.validate_changeset/2","doc":"```\n# Let's play a guessing game!\ndefmodule CoinGuess do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @primary_key false\n  embedded_schema do\n    field(:guess, Ecto.Enum, values: [:heads, :tails])\n  end\n\n  @impl InstructorLite.Instruction\n  def validate_changeset(cs, opts) do\n    target = Keyword.fetch!(opts, :extra)\n\n    case Ecto.Changeset.fetch_field!(cs, :guess) do\n      ^target -> cs\n      _ -> Ecto.Changeset.add_error(cs, :guess, \"Wrong! Try again\")\n    end\n  end\nend\n```","ref":"InstructorLite.Instruction.html#c:validate_changeset/2-example"},{"type":"module","title":"InstructorLite.Adapters.Anthropic","doc":"[Anthropic](https://docs.anthropic.com/en/home) adapter.\n\nThis adapter is implemented using the [Messages API](https://docs.anthropic.com/en/api/messages) and [function calling](https://docs.anthropic.com/en/docs/build-with-claude/tool-use).","ref":"InstructorLite.Adapters.Anthropic.html"},{"type":"module","title":"Params - InstructorLite.Adapters.Anthropic","doc":"`params` argument should be shaped as a [Create message request body](https://docs.anthropic.com/en/api/messages).","ref":"InstructorLite.Adapters.Anthropic.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.Anthropic","doc":"```\nInstructorLite.instruct(%{\n    messages: [%{role: \"user\", content: \"John is 25yo\"}],\n    model: \"claude-3-5-sonnet-20240620\",\n    metadata: %{user_id: \"3125\"}\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Anthropic,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :anthropic_key)]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.Anthropic.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nAlso specifies default `claude-3-5-sonnet-20240620` model and 1024 `max_tokens` if not provided by a user.","ref":"InstructorLite.Adapters.Anthropic.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.parse_response/2","doc":"Parse API response.\n\nCan return:\n  * `{:ok, parsed_object}` on success.\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.Anthropic.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.\n\nThe error is represented as an erroneous `tool_result`.","ref":"InstructorLite.Adapters.Anthropic.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.Anthropic.send_request/2","doc":"Make request to Anthropic API","ref":"InstructorLite.Adapters.Anthropic.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.Anthropic.send_request/2","doc":"* `:api_key` (`t:String.t/0`) - Required. Anthropic API key\n\n* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - API endpoint to use for sending requests The default value is `\"https://api.anthropic.com/v1/messages\"`.\n\n* `:version` (`t:String.t/0`) - Anthropic [API version](https://docs.anthropic.com/en/api/versioning) The default value is `\"2023-06-01\"`.","ref":"InstructorLite.Adapters.Anthropic.html#send_request/2-options"},{"type":"module","title":"InstructorLite.Adapters.ChatCompletionsCompatible","doc":"Adapter for Chat Completions-compatible API endpoints, such as [OpenAI](https://platform.openai.com/docs/api-reference/chat) or [Grok](https://docs.x.ai/docs/api-reference#chat-completions).\n\nThis adapter uses [structured outputs](https://platform.openai.com/docs/guides/structured-outputs/structured-outputs).","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html"},{"type":"module","title":"Params - InstructorLite.Adapters.ChatCompletionsCompatible","doc":"`params` argument should be shaped as a [Create chat completion request body](https://platform.openai.com/docs/api-reference/chat/create).","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.ChatCompletionsCompatible","doc":"```\nInstructorLite.instruct(%{\n    messages: [%{role: \"user\", content: \"John is 25yo\"}],\n    model: \"gpt-4o-mini\",\n    service_tier: \"default\"\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.ChatCompletionsCompatible,\n  adapter_context: [\n    api_key: Application.fetch_env!(:instructor_lite, :openai_key),\n    url: \"https://api.openai.com/v1/chat/completions\"\n  ]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.ChatCompletionsCompatible.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nAlso specifies default `gpt-4o-mini` model if not provided by a user.","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.ChatCompletionsCompatible.parse_response/2","doc":"Parse chat completion endpoint response.\n\nCan return:\n  * `{:ok, parsed_json}` on success.\n  * `{:error, :refusal, reason}` on [refusal](https://platform.openai.com/docs/guides/structured-outputs/refusals).\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.ChatCompletionsCompatible.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.ChatCompletionsCompatible.send_request/2","doc":"Make request to API.","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.ChatCompletionsCompatible.send_request/2","doc":"* `:api_key` (`t:String.t/0`) - Required. API key\n\n* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - API endpoint to use for sending requests The default value is `\"https://api.openai.com/v1/chat/completions\"`.","ref":"InstructorLite.Adapters.ChatCompletionsCompatible.html#send_request/2-options"},{"type":"module","title":"InstructorLite.Adapters.Gemini","doc":"[Gemini](https://ai.google.dev/gemini-api) adapter.\n\nThis adapter is implemented using [Text generation](https://ai.google.dev/gemini-api/docs/text-generation) endpoint configured for [structured output](https://ai.google.dev/gemini-api/docs/structured-output?lang=rest#supply-schema-in-config)","ref":"InstructorLite.Adapters.Gemini.html"},{"type":"module","title":"Params - InstructorLite.Adapters.Gemini","doc":"`params` argument should be shaped as a [`models.GenerateBody` request body](https://ai.google.dev/api/generate-content#request-body)","ref":"InstructorLite.Adapters.Gemini.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.Gemini","doc":"```\nInstructorLite.instruct(\n  %{contents: [%{role: \"user\", parts: [%{text: \"John is 25yo\"}]}]},\n  response_model: %{name: :string, age: :integer},\n  json_schema: %{\n    type: \"object\",\n    required: [:age, :name],\n    properties: %{name: %{type: \"string\"}, age: %{type: \"integer\"}}\n  },\n  adapter: InstructorLite.Adapters.Gemini,\n  adapter_context: [\n    model: \"gemini-1.5-flash-8b\",\n    api_key: Application.fetch_env!(:instructor_lite, :gemini_key)\n  ]\n)\n{:ok, %{name: \"John\", age: 25}}\n```\n\n> #### Specifying model {: .tip}\n>\n> Note how, unlike other adapters, the Gemini adapter expects `model` under `adapter_context`. \n\n> #### JSON Schema {: .warning}\n>\n> Gemini's idea of JSON Schema is [quite different](https://ai.google.dev/api/generate-content#generationconfig) from other major models, so `InstructorLite.JSONSchema` won't help you even for simple cases. Luckily, the Gemini API provides detailed errors for invalid schemas.","ref":"InstructorLite.Adapters.Gemini.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.Gemini.initial_prompt/2","doc":"Puts `systemInstruction` and updates `generationConfig` in `params` with prompt based on `json_schema` and `notes`.","ref":"InstructorLite.Adapters.Gemini.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.Gemini.parse_response/2","doc":"Parse text generation endpoint response.\n\nCan return:\n  * `{:ok, parsed_json}` on success.\n  * `{:error, :refusal, prompt_feedback}` if [request was blocked](https://ai.google.dev/api/generate-content#generatecontentresponse).\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.Gemini.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.Gemini.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.","ref":"InstructorLite.Adapters.Gemini.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.Gemini.send_request/2","doc":"Make request to Gemini API","ref":"InstructorLite.Adapters.Gemini.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.Gemini.send_request/2","doc":"* `:api_key` (`t:String.t/0`) - Required. Gemini API key\n\n* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - API endpoint to use for sending requests The default value is `\"https://generativelanguage.googleapis.com/v1beta/models/{model}:generateContent\"`.\n\n* `:model` (`t:String.t/0`) - Gemini [model](https://ai.google.dev/gemini-api/docs/models/gemini) The default value is `\"gemini-1.5-flash-8b\"`.","ref":"InstructorLite.Adapters.Gemini.html#send_request/2-options"},{"type":"module","title":"InstructorLite.Adapters.Llamacpp","doc":"[LLaMA.cpp HTTP Server](https://github.com/ggerganov/llama.cpp/tree/master/examples/server) adapter.\n\nThis adapter is implemented using llama-specific [completion](https://github.com/ggerganov/llama.cpp/tree/master/examples/server#post-completion-given-a-prompt-it-returns-the-predicted-completion) endpoint.","ref":"InstructorLite.Adapters.Llamacpp.html"},{"type":"module","title":"Params - InstructorLite.Adapters.Llamacpp","doc":"`params` argument should be shaped as a [completion request body](https://github.com/ggerganov/llama.cpp/tree/master/examples/server#post-completion-given-a-prompt-it-returns-the-predicted-completion)","ref":"InstructorLite.Adapters.Llamacpp.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.Llamacpp","doc":"```\nInstructorLite.instruct(%{\n    prompt: \"John is 25yo\",\n    temperature: 0.8\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.Llamacpp,\n  adapter_context: [url: \"http://localhost:8000/completion\"]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.Llamacpp.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nIt uses `json_schema` and `system_prompt` request parameters.","ref":"InstructorLite.Adapters.Llamacpp.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.parse_response/2","doc":"Parse API response.\n\nCan return:\n  * `{:ok, parsed_json}` on success.\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.Llamacpp.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.","ref":"InstructorLite.Adapters.Llamacpp.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.Llamacpp.send_request/2","doc":"Make request to llamacpp HTTP server.","ref":"InstructorLite.Adapters.Llamacpp.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.Llamacpp.send_request/2","doc":"* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - Required. API endpoint to use, for example `http://localhost:8000/completion`","ref":"InstructorLite.Adapters.Llamacpp.html#send_request/2-options"},{"type":"module","title":"InstructorLite.Adapters.OpenAI","doc":"[OpenAI](https://platform.openai.com/docs/overview) adapter. \n\nThis adapter is implemented using\n[responses](https://platform.openai.com/docs/api-reference/responses) endpoint\nand [structured\noutputs](https://platform.openai.com/docs/guides/structured-outputs/structured-outputs).","ref":"InstructorLite.Adapters.OpenAI.html"},{"type":"module","title":"Params - InstructorLite.Adapters.OpenAI","doc":"`params` argument should be shaped as a [Create model response request\nbody](https://platform.openai.com/docs/api-reference/responses/create).","ref":"InstructorLite.Adapters.OpenAI.html#module-params"},{"type":"module","title":"Example - InstructorLite.Adapters.OpenAI","doc":"```\nInstructorLite.instruct(%{\n    input: [%{role: \"user\", content: \"John is 25yo\"}],\n    model: \"gpt-4o-mini\",\n    service_tier: \"default\"\n  },\n  response_model: %{name: :string, age: :integer},\n  adapter: InstructorLite.Adapters.OpenAI,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %{name: \"John\", age: 25}}\n```","ref":"InstructorLite.Adapters.OpenAI.html#module-example"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.initial_prompt/2","doc":"Updates `params` with prompt based on `json_schema` and `notes`.\n\nIt uses `instructions` parameter for system prompt.\n\nAlso specifies default `gpt-4o-mini` model if not provided by a user.","ref":"InstructorLite.Adapters.OpenAI.html#initial_prompt/2"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.parse_response/2","doc":"Parse chat completion endpoint response.\n\nCan return:\n  * `{:ok, parsed_json}` on success.\n  * `{:error, :refusal, reason}` on [refusal](https://platform.openai.com/docs/guides/structured-outputs/refusals).\n  * `{:error, :unexpected_response, response}` if response is of unexpected shape.","ref":"InstructorLite.Adapters.OpenAI.html#parse_response/2"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.retry_prompt/5","doc":"Updates `params` with prompt for retrying a request.\n\nIf the initial request was made with conversation state (enabled by\ndefault), it will drop previous chat messages from the request and specify\n`previous_response_id` instead. If conversation state is disabled, it will\nappend new messages to the previous `input` the same way chat completions-based\nadapters do.","ref":"InstructorLite.Adapters.OpenAI.html#retry_prompt/5"},{"type":"function","title":"InstructorLite.Adapters.OpenAI.send_request/2","doc":"Make request to OpenAI API.","ref":"InstructorLite.Adapters.OpenAI.html#send_request/2"},{"type":"function","title":"Options - InstructorLite.Adapters.OpenAI.send_request/2","doc":"* `:api_key` (`t:String.t/0`) - Required. OpenAI API key\n\n* `:http_client` (`t:atom/0`) - Any module that follows `Req.post/2` interface The default value is `Req`.\n\n* `:http_options` (`t:keyword/0`) - Options passed to `http_client.post/2` The default value is `[receive_timeout: 60000]`.\n\n* `:url` (`t:String.t/0`) - API endpoint to use for sending requests The default value is `\"https://api.openai.com/v1/responses\"`.","ref":"InstructorLite.Adapters.OpenAI.html#send_request/2-options"},{"type":"extras","title":"InstructorLite","doc":"# InstructorLite\n\n[![Instructor version](https://img.shields.io/hexpm/v/instructor_lite.svg)](https://hex.pm/packages/instructor_lite)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/instructor_lite/)\n[![Hex Downloads](https://img.shields.io/hexpm/dt/instructor_lite)](https://hex.pm/packages/instructor_lite)\n[![GitHub stars](https://img.shields.io/github/stars/martosaur/instructor_lite.svg)](https://github.com/martosaur/instructor_lite/stargazers)\n[![CI](https://github.com/martosaur/instructor_lite/actions/workflows/ci.yml/badge.svg)](https://github.com/martosaur/instructor_lite/actions/workflows/ci.yml)\n\nStructured prompting for LLMs. InstructorLite is a fork, spiritual successor and almost an entire rewrite of [instructor_ex](https://github.com/thmsmlr/instructor_ex) library.\n \nThe Instructor is useful for coaxing an LLM to return JSON that maps to an Ecto schema that you provide, rather than the default unstructured text output. If you define your own validation logic, Instructor can automatically retry prompts when validation fails (returning natural language error messages to the LLM, to guide it when making corrections).","ref":"readme.html"},{"type":"extras","title":"Why Lite - InstructorLite","doc":"InstructorLite is designed to be:\n1. **Lean**. It does so little it makes you question if you should just write your own version!\n2. **Composable**. Almost everything it does can be overridden or extended.\n3. **Magic-free**. It doesn't hide complexity behind one line function calls, but does its best to provide you with enough information to understand what's going on.\n\nInstructorLite is tested to be compatible with the following providers:\n[OpenAI](https://openai.com/api/), [Anthropic](https://www.anthropic.com/),\n[Gemini](https://ai.google.dev/), [Grok](https://x.ai/) and\n[Llamacpp](https://github.com/ggerganov/llama.cpp).","ref":"readme.html#why-lite"},{"type":"extras","title":"Features - InstructorLite","doc":"InstructorLite can be boiled down to these features:\n1. It provides a very simple function for generating JSON schema from Ecto schema.\n2. It facilitates generating prompts, calling LLMs, casting and validating responses, including retrying prompts when validation fails.\n3. It holds knowledge of major LLM providers' API interfaces with adapters.\n\nAny of the features above can be used independently.","ref":"readme.html#features"},{"type":"extras","title":"Usage - InstructorLite","doc":"Define an instruction, which is a normal Ecto schema with an extra `use Instructor.Instruction` call.\n\n```elixir\ndefmodule UserInfo do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n  \n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:age, :integer)\n  end\nend\n```\n\nNow let's use `InstructorLite.instruct/2` to fill the schema from unstructured text:\n\n<!-- tabs-open -->","ref":"readme.html#usage"},{"type":"extras","title":"OpenAI - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    input: [\n      %{role: \"user\", content: \"John Doe is forty-two years old\"}\n    ]\n  },\n  response_model: UserInfo,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```","ref":"readme.html#openai"},{"type":"extras","title":"Anthropic - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"John Doe is forty-two years old\"}\n    ]\n  },\n  response_model: UserInfo,\n  adapter: InstructorLite.Adapters.Anthropic,\n  adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :anthropic_key)]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```","ref":"readme.html#anthropic"},{"type":"extras","title":"Llamacpp - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    prompt: \"John Doe is forty-two years old\"\n  },\n  response_model: UserInfo,\n  adapter: InstructorLite.Adapters.Llamacpp,\n  adapter_context: [url: Application.fetch_env!(:instructor_lite, :llamacpp_url)]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```","ref":"readme.html#llamacpp"},{"type":"extras","title":"Gemini - InstructorLite","doc":"```elixir\niex> InstructorLite.instruct(%{\n    contents: [\n      %{\n        role: \"user\",\n        parts: [%{text: \"John Doe is forty-two years old\"}]\n      }\n    ]\n  },\n  response_model: UserInfo,\n  json_schema: %{\n    type: \"object\",\n    required: [:age, :name],\n    properties: %{name: %{type: \"string\"}, age: %{type: \"integer\"}},\n  },\n  adapter: InstructorLite.Adapters.Gemini,\n  adapter_context: [\n    api_key: Application.fetch_env!(:instructor_lite, :gemini_key)\n  ]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```","ref":"readme.html#gemini"},{"type":"extras","title":"Grok - InstructorLite","doc":"Grok API is compatible with OpenAI Chat Completions endpoint, so we can use\nthe `ChatCompletionsCompatible` adapter with Grok's `url` and `model_name`\n\n```elixir\niex> InstructorLite.instruct(%{\n    model: \"grok-3-latest\",\n    messages: [\n      %{role: \"user\", content: \"John Doe is forty-two years old\"}\n    ]\n  },\n  response_model: UserInfo,\n  adapter: InstructorLite.Adapters.ChatCompletionsCompatible,\n  adapter_context: [\n    url: \"https://api.x.ai/v1/chat/completions\",\n    api_key: Application.fetch_env!(:instructor_lite, :grok_key)\n  ]\n)\n{:ok, %UserInfo{name: \"John Doe\", age: 42}}\n```\n\n<!-- tabs-close -->","ref":"readme.html#grok"},{"type":"extras","title":"Configuration - InstructorLite","doc":"InstructorLite _does not_ access the application environment for configuration options like adapter or API key. Instead, they're passed as options when needed. Note that different adapters may require different options, so make sure to check their documentation.","ref":"readme.html#configuration"},{"type":"extras","title":"Installation - InstructorLite","doc":"In your mix.exs, add `:instructor_lite` to your list of dependencies:\n\n```elixir\ndef deps do\n  [\n    {:instructor_lite, \"~> 1.0.0\"}\n  ]\nend\n```\n\nOptionally, include `Req` HTTP client (used by default) and Jason (for Elixir older than 1.18):\n\n```elixir\ndef deps do\n  [\n    {:req, \"~> 0.5 or ~> 1.0\"},\n    {:jason, \"~> 1.4\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Migrating from Instructor","doc":"# Migrating from Instructor\n\n`InstructorLite` is not meant as a drop-in replacement for `Instructor`, so there are a number of things to consider when making a decision to switch.\n\n✅ You might want to switch if:\n* You need more customization (e.g. using `Tesla` instead of `Req`)\n* You want to have fine-grained control over side-effects (e.g. delay executing requests to use [Batch API](https://platform.openai.com/docs/guides/batch))\n* You plan to tinker with the library's source code or write your own adapters\n* You've run into issues and and are on the path to your own fork\n\n❌ You might not want to switch if:\n* Instructor already works for you and you want things to continue to \"just work\"\n* You need streaming capabilities","ref":"migrating_from_instructor.html"},{"type":"extras","title":"Migrating Schemas - Migrating from Instructor","doc":"Instructor works with any Ecto schema and optionally allows you to use `Instructor.Validator` behaviour for the `Instructor.Validator.validate_changeset/2` callback.\n\nInstructorLite can also work with any Ecto schema, but it doesn't use the `@doc` attribute for semantic description. Instead, you can `use InstructorLite.Instruction` and define the `@notes` attribute. The `c:InstructorLite.Instruction.validate_changeset/2` callback is still here, but it's always of arity 2.\n\nAs a result, you don't have to add a docstring in your release.\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-schemas"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"","ref":"migrating_from_instructor.html#instructor"},{"type":"extras","title":"Field Descriptions: - Migrating from Instructor","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n\n```elixir\n# mix.exs\ndef project do\n  # ...\n  releases: [\n    myapp: [\n      strip_beams: [keep: [\"Docs\"]]\n    ]\n  ]\nend\n```","ref":"migrating_from_instructor.html#field-descriptions"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"","ref":"migrating_from_instructor.html#instructorlite"},{"type":"extras","title":"Field Descriptions: - Migrating from Instructor","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl InstructorLite.Instruction\n  def validate_changeset(changeset, _opts) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#field-descriptions-1"},{"type":"extras","title":"Migrating Instructor.chat_completions/2 - Migrating from Instructor","doc":"`Instructor.chat_completion/2` works primarily with `params`, but also accepts an optional `config` as the second argument. Some parameters are passed to the LLM and some control Instructor's own behavior.\n\nThe InstructorLite equivalent function is `InstructorLite.instruct/2`. It also accepts `params` as the first argument, but many params were moved to the second `opts` argument. Read more about this in [Key Concepts](`m:InstructorLite#key-concepts`).\n\nIn addition, InstructorLite does not access application configuration, so API keys are typically passed through `opts`.\n\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-instructor-chat_completions-2"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\nInstructor.chat_completion(%{\n  model: \"gpt-3.5-turbo\",\n  response_model: SpamPrediction,\n  messages: [\n    %{\n      role: \"user\",\n      content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n    }\n  ]\n})\n{:ok, %SpamPrediction{class: :spam, score: 0.999}}\n```\n\n```elixir\n# config.exs\nconfig :instructor, \n  adapter: Instructor.Adapters.OpenAI,\n  openai: [api_key: \"my_secret_key\"]\n```","ref":"migrating_from_instructor.html#instructor-1"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\nInstructorLite.instruct(%{\n    input: [\n      %{\n        role: \"user\",\n        content: \"Classify the following text: Hello, I am a Nigerian prince and I would like to give you $1,000,000.\"\n      }\n    ],\n    model: \"gpt-4o-mini\"\n  },\n  response_model: SpamPrediction,\n  adapter: InstructorLite.Adapters.OpenAI,\n  adapter_context: [api_key: \"my_secret_key\"]\n)\n{:ok, %SpamPrediction{class: :spam, score: 0.999}}\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#instructorlite-1"},{"type":"extras","title":"Migrating Custom Ecto Types - Migrating from Instructor","doc":"`Instructor.EctoType` allows you to define a way to dump a custom Ecto type to JSON schema using the `Instructor.EctoType.to_json_schema/0` callback.\n\nInstructorLite lowers the bar for the built-in Ecto-to-JSON-schema converter. It only works for basic Ecto types and may not be compatible with all LLMs. As soon as basic JSON schema is not enough for you, you should define the schema manually using the `c:InstructorLite.Instruction.json_schema/0` callback.\n\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-custom-ecto-types"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n  use Instructor.EctoType\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def to_json_schema(),\n    do: %{\n      type: \"string\",\n      description: \"A CSV representation of a data table\"\n    }\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @doc \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\nend\n```","ref":"migrating_from_instructor.html#instructor-2"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\n  \n  @impl InstructorLite.Instruction\n  def json_schema do\n    %{\n      type: \"object\",\n      description: \"\",\n      title: \"Database\",\n      required: [:tables],\n      \"$defs\": %{\n        \"DataFrame\" => %{\n          type: \"object\",\n          description: \"\",\n          title: \"DataFrame\",\n          required: [:data, :name],\n          properties: %{data: %{type: \"string\"}, name: %{type: \"string\"}},\n          additionalProperties: false\n        }\n      },\n      properties: %{tables: %{type: \"array\", items: %{\"$ref\": \"#/$defs/DataFrame\"}}},\n      additionalProperties: false\n    }\n  end\nend\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#instructorlite-2"},{"type":"extras","title":"Migrating Instructor.validate_with_llm/2 - Migrating from Instructor","doc":"`Instructor.Validator.validate_with_llm/2` is a nifty function that allows you to use LLM to validate your changeset.\n\nInstructorLite doesn't have this function. Instead, you can write your own thin helper function.\n\n<!-- tabs-open -->","ref":"migrating_from_instructor.html#migrating-instructor-validate_with_llm-2"},{"type":"extras","title":"Instructor - Migrating from Instructor","doc":"```elixir\ndefmodule QuestionAnswer do\n  use Ecto.Schema\n  use Instructor.Validator\n\n  @primary_key false\n  embedded_schema do\n    field(:question, :string)\n    field(:answer, :string)\n  end\n\n  @impl true\n  def validate_changeset(changeset) do\n    changeset\n    |> validate_with_llm(:answer, \"Do not say anything objectionable\")\n  end\nend\n```\n\n```elixir\n%QuestionAnswer{}\n|> Instructor.cast_all(%{\n  question: \"What is the meaning of life?\",\n  answer: \"Sex, drugs, and rock'n roll\"\n})\n|> QuestionAnswer.validate_changeset()\n\n#Ecto.Changeset ,\n  valid?: false\n>\n```","ref":"migrating_from_instructor.html#instructor-3"},{"type":"extras","title":"InstructorLite - Migrating from Instructor","doc":"```elixir\ndefmodule QuestionAnswer do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field :question, :string\n    field :answer, :string\n  end\n\n  def changeset(data, params) do\n    data\n    |> Ecto.Changeset.cast(params, [:answer])\n    |> validate_with_llm(:answer, \"do not say anything objectionable\")\n  end\n\n  def validate_with_llm(changeset, field, rule) do\n    Ecto.Changeset.validate_change(changeset, field, fn ^field, value ->\n      %{\n        input: [\n          %{\n            role: \"system\",\n            content: \"\"\"\n            You are a world-class validation model. Capable of determining if the following value is valid for the statement, if it is not, explain why.\n            \"\"\"\n          },\n          %{\n            role: \"user\",\n            content: \"Does `#{value}` follow the rule: #{rule}\"\n          }\n        ]\n      }\n      |> InstructorLite.instruct(\n        response_model: %{valid?: :boolean, reason: :string},\n        adapter_context: [api_key: \"my_api_token\"]\n      )\n      |> case do\n        {:ok, %{valid?: false, reason: reason}} ->\n          [{field, reason}]\n\n        _ -> []\n      end\n    end)\n  end\nend\n```\n\n```elixir\n%QuestionAnswer{question: \"What is the meaning of life?\"}\n|> QuestionAnswer.changeset(%{answer: \"Sex, drugs, and rock'n roll\"})\n\n#Ecto.Changeset ,\n  valid?: false,\n  ...\n>\n```\n\n<!-- tabs-close -->","ref":"migrating_from_instructor.html#instructorlite-3"},{"type":"extras","title":"Migrating Llamacpp Usage - Migrating from Instructor","doc":"Instructor's Llamacpp was implemented through converting JSON schema to GBNF grammar.\n\nInstructorLite offers a very basic Llamacpp adapter which passes JSON schema to the Llamacpp server. If you want to use Llamacpp with InstructorLite, you will likely need to learn how your model of choice works and write your own adapter. InstructorLite can offer you a flexible enough chassis to build upon, but doesn't have much expertise to share 🫡.","ref":"migrating_from_instructor.html#migrating-llamacpp-usage"},{"type":"extras","title":"Local Development Guide","doc":"# Local Development Guide\n\nYou should be able to run most tests out of the box:\n\n```\nmix deps.get\nmix test\n```\n\nHowever, to actually be able to query LLMs, you'll need to put your own personal API keys into a `config/config.exs`, which is ignored by Git. Start by copying `config/config.example.exs`:\n\n```\ncp config/config.example.exs config/config.exs\n```\n\nDepending on what secrets you have, you'll be able to run integration tests for different adapters. All integration tests can be found in `test/integration_test.exs`. Here's how to run the OpenAI suite:\n\n```\nmix test test/integrations/openai_test.exs\n```","ref":"local_development_guide.html"},{"type":"extras","title":"Changelog","doc":"# Changelog","ref":"changelog.html"},{"type":"extras","title":"v1.0.0 - Changelog","doc":"* OpenAI adapter is changed to use the\n  [Responses](https://platform.openai.com/docs/api-reference/responses) API.\n  The chat completions-based adapter is still available as\n  `InstructorLite.Adapters.ChatCompletionsCompatible`\n  * `Jason` dependency is now optional\n  * All adapters have been tested and verified for compatibility with reasoning models\n  * Default prompts have been refined and standardized across adapters","ref":"changelog.html#v1-0-0"},{"type":"extras","title":"Migrating to 1.0.0 - Changelog","doc":"1. If you use the `OpenAI` adapter and want to continue using the chat completions endpoint, switch to the `ChatCompletionsCompatible` adapter:\n  ```diff\n  InstructorLite.instruct(%{\n      messages: [\n        %{role: \"user\", content: \"John Doe is fourty two years old\"}\n      ]\n    },\n    response_model: %{name: :string, age: :integer},\n  -   adapter: InstructorLite.Adapters.OpenAI,\n  +   adapter: InstructorLite.Adapters.ChatCompletionsCompatible,\n    adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n  )\n  ```\n  2. If you use the `OpenAI` adapter and want to switch to the responses endpoint, update `params` to comply with the [`POST /v1/responses`](https://platform.openai.com/docs/api-reference/responses/create) interface. Most notably, the `messages` key should be `input`:\n  ```diff\n  InstructorLite.instruct(%{\n  -    messages: [\n  +    input: [\n        %{role: \"user\", content: \"John Doe is fourty two years old\"}\n      ]\n    },\n    response_model: %{name: :string, age: :integer},\n    adapter: InstructorLite.Adapters.OpenAI,\n    adapter_context: [api_key: Application.fetch_env!(:instructor_lite, :openai_key)]\n  )\n  ```","ref":"changelog.html#migrating-to-1-0-0"},{"type":"extras","title":"v0.3.0 - Changelog","doc":"* Add Gemini adapter","ref":"changelog.html#v0-3-0"},{"type":"extras","title":"v0.2.0 - Changelog","doc":"* [OpenAI] Do not overwrite `response_format` params key if provided by the user\n  * Fix `consume_response/3` enforcing all keys to present for ad-hoc Ecto schemas\n  * Add Local Development Guide","ref":"changelog.html#v0-2-0"},{"type":"extras","title":"v0.1.0 - Changelog","doc":"* Initial release","ref":"changelog.html#v0-1-0"},{"type":"extras","title":"Text Classification","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Text Classification\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, \"~> 1.0\"},\n    {:req, \"~> 0.5\"}\n  ]\n)\n```","ref":"text-classification.html"},{"type":"extras","title":"Motivation - Text Classification","doc":"Text classification is a common task in NLP and broadly applicable across software. Whether it be spam detection or support ticket categorization, NLP is at the core. Historically, this required training custom, bespoke models that required collecting thousands of pre-labeled examples. With LLMs, a lot of this knowledge is already encoded into the model. With proper instruction and guiding the output to a known set of classifications using GPT, you can be up and running with a text classification model in no time.\n\nHell, you can even use instructor to help generate the training set to train your own more efficient model. But let's not get ahead of ourselves, there's more on that later in the tutorials.","ref":"text-classification.html#motivation"},{"type":"extras","title":"Setup - Text Classification","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environmental variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"text-classification.html#setup"},{"type":"extras","title":"Binary Text Classification - Text Classification","doc":"Spam detection is a classic example of binary text classification. It's as simple as returning a true / false of whether an example is in the class. This is pretty trivial to implement in instructor.\n\n````elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"","ref":"text-classification.html#binary-text-classification"},{"type":"extras","title":"Field Descriptions: - Text Classification","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset, _opts) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n\nis_spam? = fn text ->\n  InstructorLite.instruct(%{\n      model: \"gpt-4o-mini\",\n      input: [\n        %{\n          role: \"user\",\n          content: \"\"\"\n          Your purpose is to classify customer support emails as either spam or not.\n          This is for a clothing retail business.\n          They sell all types of clothing.\n  \n          Classify the following email: \n          ```\n          #{text}\n          ```\n          \"\"\"\n        }\n      ]\n    },\n    response_model: SpamPrediction,\n    max_retries: 1,\n    adapter_context: [api_key: secret_key]\n  )\nend\n\nis_spam?.(\"Hello I am a Nigerian prince and I would like to send you money\")\n````\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %SpamPrediction{class: :spam, reason: \"Common spam trope\", score: 0.95}}\n```\n\nWe don't have to stop just at a boolean inclusion, we can also easily extend this idea to multiple categories or classes that we can classify the text into. In this example, let's consider classifying support emails. We want to know whether it's a `general_inquiry`, `billing_issue`, or a `technical_issue` perhaps it rightly fits in multiple classes. This can be useful if we want to cc' specialized support agents when intersecting customer issues occur\n\nWe can leverage `Ecto.Enum` to define a schema that restricts the LLM output to be a list of those values. We can also provide a `@notes` description to help guide the LLM with the semantic understanding of what these classifications ought to represent.\n\n```elixir\ndefmodule EmailClassifications do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  A classification of a customer support email.\n\n  technical_issue - whether the user is having trouble accessing their account\n  billing_issue - whether the customer is having trouble managing their billing or credit card\n  general_inquiry - all other issues\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:tags, {:array, Ecto.Enum},\n      values: [:general_inquiry, :billing_issue, :technical_issue]\n    )\n  end\nend\n\nclassify_email = fn text ->\n  {:ok, %{tags: result}} =\n    InstructorLite.instruct(%{\n        input: [\n          %{\n            role: \"user\",\n            content: \"Classify the following text: #{text}\"\n          }\n        ]\n      },\n      response_model: EmailClassifications,\n      adapter_context: [api_key: secret_key]\n    )\n\n  result\nend\n\nclassify_email.(\"My account is locked and I can't access my billing info.\")\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[:technical_issue, :billing_issue]\n```\n\n<!-- livebook:{\"offset\":4672,\"stamp\":{\"token\":\"XCP.WCL-sPHDy6D6Ffcx4wLTdEsSx7q-u1e24TeLgApYlMLmNC-oEIa4NvZ_wkFaqUQn9DA0lk7Wk1qgpS2zbp-yf39li8P5mipHQKfexcUdtFEPEIoUgA\",\"version\":2}} -->","ref":"text-classification.html#field-descriptions"},{"type":"extras","title":"Computer Vision - Extracting Data from Images","doc":"<!-- livebook:{\"file_entries\":[{\"name\":\"shopify-screenshot.png\",\"type\":\"attachment\"}],\"persist_outputs\":true} -->\n\n# Computer Vision - Extracting Data from Images\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, \"~> 1.0\"},\n    {:req, \"~> 0.5\"},\n    {:kino, \"~> 0.12.3\"}\n  ]\n)\n```","ref":"vision.html"},{"type":"extras","title":"Motivation - Computer Vision - Extracting Data from Images","doc":"In recent months, the latest AI research labs have turned LLMs into multimodal models. What this means is that they no longer just interpret text, but they can also interpret images. One example of this provided by Anthropic is the Claude 3.5 Sonnet model. With no extra work, you can now provide images into your prompts with Instructor and still do the normal structured extractions that you're used to.\n\nIn the following example, we will extract product details from a screenshot of a Shopify store.","ref":"vision.html#motivation"},{"type":"extras","title":"Setup - Computer Vision - Extracting Data from Images","doc":"In order to run code in this notebook, you need to add your Anthropic API key as an `ANTHROPIC_KEY` Livebook secret. It will then be accessible through an environment variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_ANTHROPIC_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"vision.html#setup"},{"type":"extras","title":"Example - Computer Vision - Extracting Data from Images","doc":"![](files/shopify-screenshot.png)\n\n```elixir\nimage = Kino.FS.file_path(\"shopify-screenshot.png\") |> File.read!()\nbase64_image = Base.encode64(image)\n\ndefmodule Product do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:price, :decimal)\n    field(:currency, Ecto.Enum, values: [:usd, :gbp, :eur, :cny])\n    field(:color, :string)\n  end\nend\n\n{:ok, result} =\n  InstructorLite.instruct(%{\n      messages: [\n        %{\n          role: \"user\",\n          content: [\n            %{type: \"text\", text: \"What is the product details of the following image?\"},\n            %{type: \"image\", source: %{data: base64_image, type: \"base64\", media_type: \"image/png\"}}\n          ]\n        }\n      ]\n    },\n    adapter: InstructorLite.Adapters.Anthropic,\n    response_model: Product,\n    adapter_context: [api_key: secret_key]\n  )\n\nresult\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Product{\n  name: \"Thomas Wooden Railway Thomas The Tank Engine\",\n  price: Decimal.new(\"33.0\"),\n  currency: :usd,\n  color: \"blue\"\n}\n```\n\n<!-- livebook:{\"offset\":2201,\"stamp\":{\"token\":\"XCP.1m6Fa57G9N3D9FsVzhEdQ8Re8DN_uJ_FkUyurSWqDTzWrMLSt-ycJi9oEzMcJ5w2nY-W5Bnvp174EnADhZBw6cc4fiJWmlsjgWY5-xkiPe_M2_MI044NgQ\",\"version\":2}} -->","ref":"vision.html#example"},{"type":"extras","title":"Converting Text into Explorer DataFrames","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Converting Text into Explorer DataFrames\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, \"~> 1.0\"},\n    {:req, \"~> 0.5\"},\n    {:explorer, \"~> 0.7.2\"},\n    {:kino, \"~> 0.12.0\"},\n    {:kino_explorer, \"~> 0.1.13\"}\n  ]\n)\n```","ref":"text-to-dataframes.html"},{"type":"extras","title":"Motivation - Converting Text into Explorer DataFrames","doc":"Sometimes we come across text that can be naturally extracted into structured data. Although we don't necessarily know how many tables, and what the schemas are represented in the data. Using Instructor we can let the LLM determine the tables and schemas for us, pulling the results into a dataframe which we can then use for further analysis.","ref":"text-to-dataframes.html#motivation"},{"type":"extras","title":"Setup - Converting Text into Explorer DataFrames","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environmental variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"text-to-dataframes.html#setup"},{"type":"extras","title":"The Schema - Converting Text into Explorer DataFrames","doc":"Notice from this example we want resulting data structure to have data frames. That's not a default Ecto type so we're gonna have to do things a little different here.\n\nLuckily, Ecto allows us to define our own custom types in which we can override how the values are casted.\n\nIn our case, we're going to create a type for a table of data, which will be represented as a CSV string with respect to the LLM. Our custom type will then parse the string when it is casted by Ecto and parse the CSV into an `Explorer.DataFrame`.\n\n```elixir\ndefmodule Ecto.CSVDataFrame do\n  use Ecto.Type\n\n  def type, do: :string\n\n  def cast(csv_str) when is_binary(csv_str) do\n    df = Explorer.DataFrame.load_csv!(csv_str)\n    {:ok, df}\n  end\n\n  def cast(%Explorer.DataFrame{} = df), do: {:ok, df}\n  def cast(_), do: :error\n\n  def dump(x), do: {:ok, x}\n  def load(x), do: {:ok, x}\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Ecto.CSVDataFrame, <<70, 79, 82, 49, 0, 0, 11, ...>>, {:load, 1}}\n```\n\nNow that we have our data type, we can use it in our embedded Ecto schema, just as we would have in any other example. The built-in `InstructorLite.JSONSchema` module won't help us here, as it doesn't know how to convert a custom Ecto type, so we'll craft our own schema in the `json_schema` callback. If everything works correctly, we should get some data frames out.\n\n```elixir\ndefmodule Database do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"\n  The extracted database will contain one or more tables with data as csv formatted with ',' delimiters\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    embeds_many :tables, DataFrame, primary_key: false do\n      field(:name, :string)\n      field(:data, Ecto.CSVDataFrame)\n    end\n  end\n\n  @impl InstructorLite.Instruction\n  def json_schema() do\n    %{\n      type: \"object\",\n      description: \"\",\n      title: \"Database\",\n      required: [:tables],\n      \"$defs\": %{\n        \"DataFrame\" => %{\n          type: \"object\",\n          description: \"\",\n          title: \"DataFrame\",\n          required: [:data, :name],\n          properties: %{data: %{type: \"string\"}, name: %{type: \"string\"}},\n          additionalProperties: false\n        }\n      },\n      properties: %{tables: %{type: \"array\", items: %{\"$ref\": \"#/$defs/DataFrame\"}}},\n      additionalProperties: false\n    }\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, Database, <<70, 79, 82, 49, 0, 0, 18, ...>>, {:json_schema, 0}}\n```\n\nPerfect. And now let's test it out on some example text.\n\n```elixir\nextract_df = fn text ->\n  InstructorLite.instruct(%{\n      model: \"gpt-4o-mini\",\n      input: [\n        %{\n          role: \"system\",\n          content: \"\"\"\n          Map this data into one or more dataframes and correctly define the columns and data\n          \"\"\"\n        },\n        %{\n          role: \"user\",\n          content: \"#{text}\"\n        }\n      ]\n    },\n    response_model: Database,\n    adapter_context: [api_key: secret_key]\n  )\nend\n\n{:ok, db} =\n  extract_df.(\"\"\"\n    My name is John and I am 25 years old. I live in\n    New York and I like to play basketball. His name is\n    Mike and he is 30 years old. He lives in San Francisco\n    and he likes to play baseball. Sarah is 20 years old\n    and she lives in Los Angeles. She likes to play tennis.\n    Her name is Mary and she is 35 years old.\n    She lives in Chicago.\n    \n    On one team 'Tigers' the captain is John and there are 12 players.\n    On the other team 'Lions' the captain is Mike and there are 10 players.\n  \"\"\")\n\nKino.Layout.tabs(\n  for table <- db.tables do\n    {table.name, table.data}\n  end\n)\n```\n\n<!-- livebook:{\"offset\":4584,\"stamp\":{\"token\":\"XCP.Q2Z-FmkgAonxmyIqoldwCJ6FVZEZjMt07JUB5ZF2sx2tqhP2RpNipmNf5d1KS-RUSMkh9NQpqt-0jEz3WrKkji0QsQ1KPemSLbwLgVcYQ-aWuaQufg\",\"version\":2}} -->","ref":"text-to-dataframes.html#the-schema"},{"type":"extras","title":"Using OpenAI Batch API","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Using OpenAI Batch API\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, \"~> 1.0\"},\n    {:req, \"~> 0.5\"},\n    {:multipart, \"0.4.0\"}\n  ]\n)\n```","ref":"batch-api.html"},{"type":"extras","title":"Motivation - Using OpenAI Batch API","doc":"OpenAI offers asynchronous [Batch API](https://platform.openai.com/docs/guides/batch) at a 50% discount. This can make a huge difference on the bill if we have a lot of workload that is done in the background and isn't time-sensitive.","ref":"batch-api.html#motivation"},{"type":"extras","title":"Setup - Using OpenAI Batch API","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an `OPENAI_KEY` Livebook secret. It will then be accessible through an environment variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"batch-api.html#setup"},{"type":"extras","title":"Classifying All the Emails - Using OpenAI Batch API","doc":"We'll use our classic email classification example, but this time for a bunch of emails at once.\n\n```elixir\ndefmodule SpamPrediction do\n  use Ecto.Schema\n  use InstructorLite.Instruction\n\n  @notes \"\"\"","ref":"batch-api.html#classifying-all-the-emails"},{"type":"extras","title":"Field Descriptions: - Using OpenAI Batch API","doc":"- class: Whether or not the email is spam.\n  - reason: A short, less than 10 word rationalization for the classification.\n  - score: A confidence score between 0.0 and 1.0 for the classification.\n  \"\"\"\n  @primary_key false\n  embedded_schema do\n    field(:class, Ecto.Enum, values: [:spam, :not_spam])\n    field(:reason, :string)\n    field(:score, :float)\n  end\n\n  @impl true\n  def validate_changeset(changeset, _opts) do\n    changeset\n    |> Ecto.Changeset.validate_number(:score,\n      greater_than_or_equal_to: 0.0,\n      less_than_or_equal_to: 1.0\n    )\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, SpamPrediction, <<70, 79, 82, 49, 0, 0, 19, ...>>, {:validate_changeset, 2}}\n```\n\n```elixir\nopts = [\n  response_model: SpamPrediction,\n  adapter: InstructorLite.Adapters.OpenAI\n]\n\nprompts =\n  [\n    \"Hello I am a Nigerian prince and I would like to send you money\",\n    \"Hello I know we never met but I got access to your webcam and have some spicy footage\",\n    \"Elixir Radar Newsletter. Issue #420\"\n  ]\n  |> Enum.map(fn text ->\n    InstructorLite.prepare_prompt(%{\n        model: \"gpt-4o-mini\",\n        input: [%{\n          role: \"user\",\n          content: \"Classify first paragraph of an email:\\n#{text}\"\n        }]\n      },\n      opts\n    )\n  end)\n  |> Enum.with_index()\n\njsonl =\n  prompts\n  |> Enum.map(fn {prompt, index} ->\n    Jason.encode!(%{custom_id: \"#{index}\", method: \"POST\", url: \"/v1/responses\", body: prompt})\n  end)\n  |> Enum.join(\"\\n\")\n\nmultipart =\n  Multipart.new()\n  |> Multipart.add_part(Multipart.Part.text_field(\"batch\", \"purpose\"))\n  |> Multipart.add_part(\n    Multipart.Part.file_content_field(\"emails.jsonl\", jsonl, :file, filename: \"emails.jsonl\")\n  )\n\nauth_headers = [{\"Authorization\", \"Bearer #{secret_key}\"}]\nmp_headers = [{\"Content-Type\", Multipart.content_type(multipart, \"multipart/form-data\")}]\n\n{:ok, %{body: %{\"id\" => file_id}}} =\n  Req.post(\"https://api.openai.com/v1/files\",\n    headers: auth_headers ++ mp_headers,\n    body: Multipart.body_binary(multipart)\n  )\n\n{:ok, %{body: %{\"id\" => batch_id}}} =\n  Req.post(\"https://api.openai.com/v1/batches\",\n    json: %{input_file_id: file_id, endpoint: \"/v1/responses\", completion_window: \"24h\"},\n    headers: auth_headers\n  )\n\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```\n\nNow we need to wait a little and fetch the batch result.\n\n```elixir\n{:ok, %{body: %{\"output_file_id\" => output_file_id}}} =\n  Req.get(\"https://api.openai.com/v1/batches/#{batch_id}\", headers: auth_headers)\n\n{:ok, %{body: body}} =\n  Req.get(\"https://api.openai.com/v1/files/#{output_file_id}/content\", headers: auth_headers)\n\nresult =\n  body\n  |> String.split(\"\\n\", trim: true)\n  |> Enum.map(&Jason.decode!/1)\n  |> Enum.zip(prompts)\n  |> Enum.map(fn {%{\"response\" => %{\"body\" => response}}, prompt} ->\n    InstructorLite.consume_response(response, prompt, response_model: SpamPrediction)\n  end)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n[\n  ok: %SpamPrediction{class: :spam, reason: \"Typical scam offer from prince.\", score: 0.95},\n  ok: %SpamPrediction{class: :spam, reason: \"Suspicious webcam claim\", score: 0.95},\n  ok: %SpamPrediction{\n    class: :not_spam,\n    reason: \"Newsletter content is not promotional\",\n    score: 0.85\n  }\n]\n```\n\n<!-- livebook:{\"offset\":4253,\"stamp\":{\"token\":\"XCP._idhRxSiScwUoVVMHrkKP07iyaWcBpp0qsI6L46lHzCHDtaEkUXv1pNoHa1W18Cp2LGOw7x4R9EJlMQ97Zryy-5-OMQ65Mru-brPSwWE3ofEEgX_oA\",\"version\":2}} -->","ref":"batch-api.html#field-descriptions"},{"type":"extras","title":"Custom Ollama Adapter","doc":"<!-- livebook:{\"persist_outputs\":true} -->\n\n# Custom Ollama Adapter\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, \"~> 1.0\"},\n    {:req, \"~> 0.5\"}\n  ]\n)\n```","ref":"custom-ollama-adapter.html"},{"type":"extras","title":"About Ollama - Custom Ollama Adapter","doc":"[Ollama](https://ollama.com/) project provides a way to run open source LLMs locally, provided you have capable hardware.\n\nTo run a model, just pick one from [the list](https://ollama.com/search) and run CLI command:\n\n```\n> ollama run llama3.1\npulling manifest\npulling 8eeb52dfb3bb... 100% ▕█████████████████████████████████████████████████████████████████████████████████████████████▏ 4.7 GB\npulling 948af2743fc7... 100% ▕█████████████████████████████████████████████████████████████████████████████████████████████▏ 1.5 KB\npulling 0ba8f0e314b4... 100% ▕█████████████████████████████████████████████████████████████████████████████████████████████▏  12 KB\npulling 56bb8bd477a5... 100% ▕█████████████████████████████████████████████████████████████████████████████████████████████▏   96 B\npulling 1a4c3c319823... 100% ▕█████████████████████████████████████████████████████████████████████████████████████████████▏  485 B\nverifying sha256 digest\nwriting manifest\nsuccess\n```\n\nThe open source models come in so many shapes and sizes that it is impractical to try to fit them all into a single adapter. But have no fear, as InstructorLite makes it simple to write one for your specific needs!","ref":"custom-ollama-adapter.html#about-ollama"},{"type":"extras","title":"Custom Adapter - Custom Ollama Adapter","doc":"For this exercise, we'll write a custom adapter for Ollama running the Llama3.1 model.\n\nTo do this, we need to create a module implementing the `InstructorLite.Adapter` behaviour, which requires four callbacks. Let's go through them one by one.\n\n<!-- livebook:{\"break_markdown\":true} -->\n\nThe first callback is `send_request/2`, which is responsible for making a call with `params` to some kind of API endpoint. Looking at the Ollama docs, [chat completion](https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-chat-completion) seems like a good endpoint to aim for, but there's no need to hardcode it in the adapter. We'll just require callers to provide it in full, including hostname and port. Since this is a model running locally, it will likely be something like `http://localhost:11434/api/chat`.\n\n```elixir\ndefmodule OllamaAdapterSendRequest do\n  def send_request(params, options) do\n    # Remember, adapter-specific options are nested under `adapter_context` key\n    url = options |> Keyword.get(:adapter_context, []) |> Keyword.fetch!(:url)\n\n    case Req.post(url, json: params) do\n      {:ok, %{status: 200, body: body}} -> {:ok, body}\n      {:ok, response} -> {:error, response}\n      {:error, reason} -> {:error, reason}\n    end\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, OllamaAdapterSendRequest, <<70, 79, 82, 49, 0, 0, 8, ...>>, {:send_request, 2}}\n```\n\nThe next callback is `initial_prompt/2`. This one enriches user-provided `params` with whatever is necessary to convince the LLM to fill the JSON schema. This part is very model-specific, so we need to consult the docs.\n\nFast forward, after reading [the model card](https://ollama.com/library/llama3.1), [the chat completion](https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-completion) endpoint documentation and some explorative testing, we see that Llama3.1 supports using tools, which is one common way to ensure JSON output, used for example in the `InstructorLite.Adapters.Anthropic` adapter. The idea is that we can pretend that our JSON schema is a function tool, that the model can \"call\". Normally, a model will then receive the result of a function call and continue dialogue, but we will be only interested in the call itself. Okay, let's implement the `initial_prompt/2` callback using tool calls\n\n```elixir\ndefmodule OllamaAdapterInitialPrompt do\n  def initial_prompt(params, opts) do\n    # This prompt will be a message with a \"system\" role\n    sys_message = [\n      %{\n        role: \"system\",\n        content: InstructorLite.Prompt.prompt(opts)\n      }\n    ]\n\n    params\n    # This adapter is model-specific, so might as well enforce it here\n    |> Map.put(:model, \"llama3.1\")\n    |> Map.put(:stream, false)\n    # The user has likely provided their own prompt, so we need to be nice and not overwrite it\n    |> Map.update(:messages, sys_message, fn msgs -> sys_message ++ msgs end)\n    # Finally, we represent our json schema as a function tool available for a call\n    |> Map.put(:tools, [\n      %{\n        type: \"function\",\n        function: %{\n          name: \"Schema\",\n          description:\n            \"Correctly extracted `Schema` with all the required parameters with correct types\",\n          parameters: Keyword.fetch!(opts, :json_schema)\n        }\n      }\n    ])\n  end\nend\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:module, OllamaAdapterInitialPrompt, <<70, 79, 82, 49, 0, 0, 10, ...>>, {:initial_prompt, 2}}\n```\n\nAt this point, we can actually try to combine both callbacks and see if we can get a response.\n\n```elixir\njson_schema = InstructorLite.JSONSchema.from_ecto_schema(%{name: :string})\nprompt = OllamaAdapterInitialPrompt.initial_prompt(%{messages: [\n  %{role: \"user\", content: \"Who was the first president of the united states?\"}\n]}, json_schema: json_schema)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%{\n  messages: [\n    %{\n      role: \"system\",\n      content: \"You're called by an Elixir application through the InstructorLite library. Your task is to understand what the application wants you to do and respond with JSON output that matches the schema. The output will be validated by the application against an Ecto schema and potentially some custom rules. You may be asked to adjust your response if it doesn't pass validation. \"\n    },\n    %{role: \"user\", content: \"Who was the first president of the united states?\"}\n  ],\n  stream: false,\n  tools: [\n    %{\n      function: %{\n        name: \"Schema\",\n        description: \"Correctly extracted `Schema` with all the required parameters with correct types\",\n        parameters: %{\n          type: \"object\",\n          title: \"root\",\n          required: [:name],\n          additionalProperties: false,\n          properties: %{name: %{type: \"string\"}}\n        }\n      },\n      type: \"function\"\n    }\n  ],\n  model: \"llama3.1\"\n}\n```\n\n```elixir\nurl = \"http://localhost:11434/api/chat\"\n{:ok, response} = OllamaAdapterSendRequest.send_request(prompt, adapter_context: [url: url])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok,\n %{\n   \"created_at\" => \"2025-06-16T16:37:25.172062Z\",\n   \"done\" => true,\n   \"done_reason\" => \"stop\",\n   \"eval_count\" => 17,\n   \"eval_duration\" => 1411956916,\n   \"load_duration\" => 36709250,\n   \"message\" => %{\n     \"content\" => \"\",\n     \"role\" => \"assistant\",\n     \"tool_calls\" => [\n       %{\"function\" => %{\"arguments\" => %{\"name\" => \"George Washington\"}, \"name\" => \"Schema\"}}\n     ]\n   },\n   \"model\" => \"llama3.1\",\n   \"prompt_eval_count\" => 243,\n   \"prompt_eval_duration\" => 4897901084,\n   \"total_duration\" => 6350061833\n }}\n```\n\nGreat! This brings us to the `parse_response/2` callback. Its job is to extract a parsed JSON object from the raw response, which we'll subsequently attempt to cast to our Ecto schema.\n\n```elixir\ndefmodule OllamaAdapterParseResponse do\n  def parse_response(response, _opts) do\n    case response do\n      %{\n        \"message\" => %{\"tool_calls\" => [%{\"function\" => %{\"arguments\" => json}}]}\n      } ->\n        {:ok, json}\n\n      other ->\n        {:error, :unexpected_response, other}\n    end\n  end\nend\n\n{:ok, params} = OllamaAdapterParseResponse.parse_response(response, [])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %{\"name\" => \"George Washington\"}}\n```\n\nThat is actually enough for a happy path! But there is one more callback that deals with cases when the output doesn't quite adhere to the schema. Imagine if we received a different response from a model:\n\n```elixir\nhypothetical = %{\"name\" => false}\nchangeset = Ecto.Changeset.cast({%{}, %{name: :string}}, hypothetical, [:name])\nerrors = InstructorLite.ErrorFormatter.format_errors(changeset)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\"name - is invalid\"\n```\n\nWe need to retry our query and relay the errors to the model, so it can correct its previous answer. `retry_callback/5` is the biggest callback of all, simply because it can potentially need so much context to do that. It needs to know what `params` were, which data we attempted to cast and which errors we saw, as well as the raw response and, of course, options. For this adapter, we will only need `params`, `errors` and raw `response`.\n\n```elixir\ndefmodule OllamaAdapterRetryPrompt do\n  def retry_prompt(params, _resp_params, errors, response, _opts) do\n    message = Map.fetch!(response, \"message\")\n\n    do_better = [\n      # We need to append model's response to the dialogue so it would know what its previous answer was\n      message,\n      %{\n        # Ollama has a `tool` message role which we are going to use\n        role: \"tool\",\n        content: InstructorLite.Prompt.validation_failed(errors)\n      }\n    ]\n\n    Map.update(params, :messages, do_better, fn msgs -> msgs ++ do_better end)\n  end\nend\n\nretry_prompt = OllamaAdapterRetryPrompt.retry_prompt(prompt, nil, errors, response, [])\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%{\n  messages: [\n    %{\n      role: \"system\",\n      content: \"You're called by an Elixir application through the InstructorLite library. Your task is to understand what the application wants you to do and respond with JSON output that matches the schema. The output will be validated by the application against an Ecto schema and potentially some custom rules. You may be asked to adjust your response if it doesn't pass validation. \"\n    },\n    %{role: \"user\", content: \"Who was the first president of the united states?\"},\n    %{\n      \"content\" => \"\",\n      \"role\" => \"assistant\",\n      \"tool_calls\" => [\n        %{\"function\" => %{\"arguments\" => %{\"name\" => \"George Washington\"}, \"name\" => \"Schema\"}}\n      ]\n    },\n    %{\n      role: \"tool\",\n      content: \"The response did not pass validation. Please try again and fix the following validation errors:\\n\\nname - is invalid\\n\"\n    }\n  ],\n  stream: false,\n  tools: [\n    %{\n      function: %{\n        name: \"Schema\",\n        description: \"Correctly extracted `Schema` with all the required parameters with correct types\",\n        parameters: %{\n          type: \"object\",\n          title: \"root\",\n          required: [:name],\n          additionalProperties: false,\n          properties: %{name: %{type: \"string\"}}\n        }\n      },\n      type: \"function\"\n    }\n  ],\n  model: \"llama3.1\"\n}\n```\n\nThis retry prompt can now be used for a `send_request/2` call. We won't do it though, to avoid gaslighting the model by suggesting its answer was incorrect.","ref":"custom-ollama-adapter.html#custom-adapter"},{"type":"extras","title":"Tying It All Together - Custom Ollama Adapter","doc":"Now we can put all these callbacks into a single module, confess to the compiler that our intention is to implement a behavior, and let the `InstructorLite` main interface handle all the logistics.\n\n```elixir\ndefmodule OllamaAdapter do\n  @behaviour InstructorLite.Adapter\n\n  @impl InstructorLite.Adapter\n  def send_request(params, options) do\n    url = options |> Keyword.get(:adapter_context, []) |> Keyword.fetch!(:url)\n\n    case Req.post(url, json: params) do\n      {:ok, %{status: 200, body: body}} -> {:ok, body}\n      {:ok, response} -> {:error, response}\n      {:error, reason} -> {:error, reason}\n    end\n  end\n\n  @impl InstructorLite.Adapter\n  def initial_prompt(params, opts) do\n    sys_message = [\n      %{\n        role: \"system\",\n        content: InstructorLite.Prompt.prompt(opts)\n      }\n    ]\n\n    params\n    |> Map.put(:model, \"llama3.1\")\n    |> Map.put(:stream, false)\n    |> Map.update(:messages, sys_message, fn msgs -> sys_message ++ msgs end)\n    |> Map.put(:tools, [\n      %{\n        type: \"function\",\n        function: %{\n          name: \"Schema\",\n          description:\n            \"Correctly extracted `Schema` with all the required parameters with correct types\",\n          parameters: Keyword.fetch!(opts, :json_schema)\n        }\n      }\n    ])\n  end\n\n  @impl InstructorLite.Adapter\n  def parse_response(response, _opts) do\n    case response do\n      %{\n        \"message\" => %{\"tool_calls\" => [%{\"function\" => %{\"arguments\" => json}}]}\n      } ->\n        {:ok, json}\n\n      other ->\n        {:error, :unexpected_response, other}\n    end\n  end\n\n  @impl InstructorLite.Adapter\n  def retry_prompt(params, _resp_params, errors, response, _opts) do\n    message = Map.fetch!(response, \"message\")\n\n    do_better = [\n      message,\n      %{\n        role: \"tool\",\n        content: InstructorLite.Prompt.validation_failed(errors)\n      }\n    ]\n\n    Map.update(params, :messages, do_better, fn msgs -> msgs ++ do_better end)\n  end\nend\n\nInstructorLite.instruct(%{\n    messages: [\n      %{role: \"user\", content: \"Dr. John Doe is forty two\"}\n    ],\n  },\n  adapter: OllamaAdapter,\n  notes: \"Trim down all people's titles\",\n  response_model: %{name: :string, age: :integer},\n  adapter_context: [\n    url: \"http://localhost:11434/api/chat\"\n  ]\n)\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n{:ok, %{name: \"John Doe\", age: 42}}\n```","ref":"custom-ollama-adapter.html#tying-it-all-together"},{"type":"extras","title":"OpenAI Advanced Features","doc":"<!-- livebook:{\"file_entries\":[{\"name\":\"shopify-screenshot.png\",\"type\":\"attachment\"}],\"persist_outputs\":true} -->\n\n# OpenAI Advanced Features\n\n```elixir\nMix.install(\n  [\n    {:instructor_lite, \"~> 1.0\"},\n    {:req, \"~> 0.5\"},\n    {:kino, \"~> 0.16\"}\n  ]\n)\n```","ref":"openai-advanced-features.html"},{"type":"extras","title":"Motivation - OpenAI Advanced Features","doc":"OpenAI has introduced a lot of features since InstructorLite was first released. Let's explore some of them and see if they can be of any help.","ref":"openai-advanced-features.html#motivation"},{"type":"extras","title":"Setup - OpenAI Advanced Features","doc":"In order to run code in this notebook, you need to add your OpenAI API key as an OPENAI_KEY Livebook secret. It will then be accessible through an environment variable.\n\n```elixir\nsecret_key = System.fetch_env!(\"LB_OPENAI_KEY\")\n:ok\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n:ok\n```","ref":"openai-advanced-features.html#setup"},{"type":"extras","title":"Image Input - OpenAI Advanced Features","doc":"![](files/shopify-screenshot.png)\n\n```elixir\nfile = Kino.FS.file_path(\"shopify-screenshot.png\") |> File.read!()\nbase64_image = \"data:image/png;base64,\" <> Base.encode64(file)\n\ndefmodule Product do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:price, :decimal)\n    field(:currency, Ecto.Enum, values: [:usd, :gbp, :eur, :cny])\n    field(:color, :string)\n  end\nend\n\n{:ok, result} =\n  InstructorLite.instruct(%{\n      model: \"gpt-4o-mini\",\n      input: [\n        %{\n          role: \"user\",\n          content: [\n            %{type: \"input_text\", text: \"What is the product details of the following image?\"},\n            %{type: \"input_image\", image_url: base64_image}\n          ]\n        }\n      ]\n    },\n    response_model: Product,\n    adapter_context: [api_key: secret_key]\n  )\n\nresult\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Product{\n  name: \"Thomas Wooden Railway Thomas the Tank Engine\",\n  price: Decimal.new(\"33.0\"),\n  currency: :usd,\n  color: \"blue\"\n}\n```","ref":"openai-advanced-features.html#image-input"},{"type":"extras","title":"Built-in Tool Use - OpenAI Advanced Features","doc":"OpenAI supports a number of built-in tools, such as web search!\n\n```elixir\ndefmodule Package do\n  use Ecto.Schema\n\n  @primary_key false\n  embedded_schema do\n    field(:name, :string)\n    field(:github_link, :string)\n    field(:latest_version, :string)\n  end\nend\n\n{:ok, result} =\n  InstructorLite.instruct(%{\n      model: \"gpt-4o-mini\",\n      tools: [%{type: \"web_search_preview\"}],\n      input: \"We want to know about: InstructorLite Hex package\"\n    },\n    response_model: Package,\n    adapter_context: [api_key: secret_key]\n  )\n\nresult\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%Package{\n  name: \"InstructorLite\",\n  github_link: \"https://github.com/martosaur/instructor_lite\",\n  latest_version: \"1.0.0\"\n}\n```","ref":"openai-advanced-features.html#built-in-tool-use"},{"type":"extras","title":"Reasoning - OpenAI Advanced Features","doc":"Reasoning models are powerful, but ultimately, reasoning doesn't affect structured output in any way. However, they can be very useful to keep track of _the reasoning_ behind the output. `InstructorLite.instruct/2` does not give you access to the raw output, but you can use lower-level functions to sidestep this.\n\n```elixir\nrequire Logger\n\nadventure_description = \"\"\"\nSomething strange is happening to your hometown of Willowshore! Nestled on the\nbanks of a river winding through the legendary Specterwood in Tian Xia’s haunted\nland of Shenmen, the people of Willowshore are no strangers to supernatural\nthreat, but the danger that comes to town on the first day of summer is unlike\nanything you’ve ever seen before. Over the four seasons to come, you and your\nfellow home-grown heroes must face evil spirits, sinister fiends, and\nfrightening curses, lest the town of Willowshore succumb to the Season of\nGhosts!\n\"\"\"\n\nopts = [\n  response_model: %{toponyms: {:array, :string}},\n  adapter_context: [api_key: secret_key]\n]\n\nparams = InstructorLite.prepare_prompt(%{\n      model: \"o4-mini\",\n      reasoning: %{effort: \"medium\", summary: \"auto\"},\n      input: [\n        %{role: \"system\", content: \"Here's an adventure description, please extract all toponyms\"},\n        %{role: \"user\", content: adventure_description}\n      ]\n  }, opts)\n\n{:ok, response} = InstructorLite.Adapters.OpenAI.send_request(params, opts)\n{:ok, result} = InstructorLite.consume_response(response, params, opts)\n\nfor %{\"summary\" => summaries} <- response[\"output\"] do\n  for %{\"type\" => \"summary_text\", \"text\" => text} <- summaries do\n    Logger.info(text)\n  end\nend\n\nresult\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n\n13:34:05.078 [info] **Extracting toponyms**\n\nThe user is interested in extracting all toponyms from a prompt about Willowshore. I’ll identify the place names, which include: Willowshore, Specterwood, Tian Xia, and Shenmen. Event names, like \"the Season of Ghosts,\" aren't toponyms. The phrase \"the banks of a river\" doesn't name the river either. \n\nI’ll ensure to include only unique entries, ultimately returning this list in a structured format as JSON: {\"toponyms\":[\"Willowshore\",\"Specterwood\",\"Tian Xia\",\"Shenmen\"]}.\n\n```\n\n<!-- livebook:{\"output\":true} -->\n\n```\n%{toponyms: [\"Willowshore\", \"Specterwood\", \"Tian Xia\", \"Shenmen\"]}\n```\n\n<!-- livebook:{\"offset\":4852,\"stamp\":{\"token\":\"XCP.s6WZuHcTI5hI4YuZjQtQOV488LZwjgSEUuzWaPfyGotNZM80UEAKeY3CtGTlJSggF1Az9UxZe39vhJ3HWk9mALIJeh6Db2bfSGIXrWsKTctJcdiAXA\",\"version\":2}} -->","ref":"openai-advanced-features.html#reasoning"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}